/***************************************************
  Code derived from the Adafruit MQTT Library Ethernet Example

  Adafruit invests time and resources providing this open source code,
  please support Adafruit and open-source hardware by purchasing
  products from Adafruit!

  Written by Alec Moore
  Derived from the code written by Limor Fried/Ladyada for Adafruit Industries.
  MIT license, all text above must be included in any redistribution
 ****************************************************/

// include libraries for controlling servos and debouncing buttons
#include <Servo.h>
#include <Bounce2.h>

// include libraries for internet functionality
#include <SPI.h>
#include "Adafruit_MQTT.h"
#include "Adafruit_MQTT_Client.h"
#include <Ethernet.h>
#include <EthernetClient.h>
#include <Dns.h>
#include <Dhcp.h>

// ethernet setup
byte mac[] = {0xDE, 0xAD, 0xBE, 0xEF, 0xFE, 0xED};
//Uncomment the following, and set to a valid ip if you don't have dhcp available.
//IPAddress iotIP (192, 168, 0, 42);
//Uncomment the following, and set to your preference if you don't have automatic dns.
//IPAddress dnsIP (8, 8, 8, 8);
//If you uncommented either of the above lines, make sure to change "Ethernet.begin(mac)" to "Ethernet.begin(mac, iotIP)" or "Ethernet.begin(mac, iotIP, dnsIP)"

// adafruit.io setup
#define AIO_SERVER      "io.adafruit.com"
#define AIO_SERVERPORT  1883
#define AIO_USERNAME    "your_user_name"
#define AIO_KEY         "your_adafruit_io_key"

// globals
EthernetClient client;
Adafruit_MQTT_Client mqtt(&client, AIO_SERVER, AIO_SERVERPORT, AIO_USERNAME, AIO_KEY);
#define halt(s) { Serial.println(F( s )); while(1);  }

// feeds
Adafruit_MQTT_Subscribe alarmTimeFeed = Adafruit_MQTT_Subscribe(&mqtt, AIO_USERNAME "/feeds/alarmtime");
Adafruit_MQTT_Subscribe alarmMelodyFeed = Adafruit_MQTT_Subscribe(&mqtt, AIO_USERNAME "/feeds/alarmmelody");
Adafruit_MQTT_Subscribe alarmToggleFeed = Adafruit_MQTT_Subscribe(&mqtt, AIO_USERNAME "/feeds/alarmtoggle");
Adafruit_MQTT_Subscribe utcTimeFeed = Adafruit_MQTT_Subscribe(&mqtt, "time/ISO-8601");

// declare objects to represent the two servos
Servo bigServo;
Servo littleServo;

// declare object to represent the button
Bounce button = Bounce();

// set global variables
int notePositions[] = {109,104,99,94,89,85,79,75};
int preHitDelay = 150;
int postHitDelay = 50;
int beaterRestAngle = 100;
int beaterHitAngle = 117;
unsigned int tempoDelay;
int tempo = 60; // bpm
int currentNote = 0;
int ledPin = 2;
int noteSequence[50] = {0,1,3,4,5};
int noteSequenceLength = 5;
unsigned long secondsSinceMidnight;
unsigned long alarmTime;
boolean alarmSet = false;
boolean alarmPlayedToday = false;
int timeZoneOffset = 0; // UK = 0, New York = -5, etc
int daylightSavingsOffset = 1;

void setup() {
  bigServo.attach(8);
  littleServo.attach(9);
  bigServo.write(90);
  littleServo.write(beaterRestAngle);
  button.attach(3, INPUT_PULLUP);
  button.interval(25);
  tempoDelay = 60000 / (tempo * 2) - preHitDelay - postHitDelay;
  pinMode(ledPin, OUTPUT);
  digitalWrite(ledPin, alarmSet);
  
  Serial.begin(115200);
  Serial.println(F("Hackspace glockenspiel alarm clock!"));
  Serial.println(F("Init the client..."));
  Ethernet.begin(mac);
  delay(1000); //give the ethernet a second to initialize
  mqtt.subscribe(&alarmTimeFeed);
  mqtt.subscribe(&alarmMelodyFeed);
  mqtt.subscribe(&alarmToggleFeed);
  mqtt.subscribe(&utcTimeFeed);
}

void loop() {
  // Ensure the connection to the MQTT server is alive (this will make the first
  // connection and automatically reconnect when disconnected).  See the MQTT_connect
  // function definition further below.
  MQTT_connect();

  // this is our 'wait for incoming subscription packets' busy subloop
  Adafruit_MQTT_Subscribe *subscription;
  while ((subscription = mqtt.readSubscription(1000))) {
    if (subscription == &alarmTimeFeed) {
      Serial.print(F("Got alarm time: "));
      Serial.println((char *)alarmTimeFeed.lastread);
      flashLEDs();
      int h1 = alarmTimeFeed.lastread[0] - '0'; // the - '0' part converts from char to int
      int h2 = alarmTimeFeed.lastread[1] - '0';
      int m1 = alarmTimeFeed.lastread[2] - '0';
      int m2 = alarmTimeFeed.lastread[3] - '0';
      unsigned long hour = 10*h1 + h2;
      unsigned long minute = 10*m1 + m2;
      alarmTime = 60*60*hour + 60*minute;
      alarmPlayedToday = (secondsSinceMidnight > alarmTime);
    }
    if (subscription == &alarmMelodyFeed) {
      Serial.print(F("Got alarm melody: "));
      Serial.println((char *)alarmMelodyFeed.lastread);
      flashLEDs();
      boolean sequenceFinished = false;
      for(int i=0; i<SUBSCRIPTIONDATALEN && !sequenceFinished; i++) {
        int thisNote = alarmMelodyFeed.lastread[i];
        if(thisNote == 0) {
          sequenceFinished = true;
          noteSequenceLength = i;
        } else {
          // convert notes from ascii codes to note numbers (c=0,d=1,e=2...C=7)
          if(thisNote == 67) { // 67 is the ascii code for 'C'
            thisNote = 7;
          } else {
            thisNote -= 99; // 99 is the ascii code for 'c'
            if(thisNote < 0 && thisNote >= -2) thisNote += 7;
          }
          if(thisNote < 0 || thisNote > 7) thisNote = -1;
          noteSequence[i] = thisNote;
        }
      }
    }
    if (subscription == &alarmToggleFeed) {
      Serial.print(F("Got alarm on/off: "));
      Serial.println((char *)alarmToggleFeed.lastread);
      alarmSet = alarmToggleFeed.lastread[0] == '1';
      alarmPlayedToday = (secondsSinceMidnight > alarmTime);
      digitalWrite(ledPin, alarmSet);
    }
    if (subscription == &utcTimeFeed) {
      Serial.print(F("Got UTC time: "));
      Serial.println((char *)utcTimeFeed.lastread);
      int h1 = utcTimeFeed.lastread[11] - '0'; // the - '0' part converts from char to int
      int h2 = utcTimeFeed.lastread[12] - '0';
      int m1 = utcTimeFeed.lastread[14] - '0';
      int m2 = utcTimeFeed.lastread[15] - '0';
      int s1 = utcTimeFeed.lastread[17] - '0';
      int s2 = utcTimeFeed.lastread[18] - '0';
      long hour = 10*h1 + h2 + timeZoneOffset + daylightSavingsOffset;
      if(hour < 0) hour += 24;
      unsigned long minute = 10*m1 + m2;
      unsigned long second = 10*s1 + s2;
      unsigned long previousTime = secondsSinceMidnight;
      secondsSinceMidnight = 60 * 60 * hour + 60 * minute + second;
      if(secondsSinceMidnight < previousTime) {
        // new day! reset alarm
        alarmPlayedToday = false;
      }
    }

    // play alarm when the time is right
    if(secondsSinceMidnight >= alarmTime && alarmSet && !alarmPlayedToday) {
      alarmPlayedToday = true;
      playMelody();
    }
  }

  // ping the server to keep the mqtt connection alive
  if(! mqtt.ping()) {
    mqtt.disconnect();
  }
}

void playMelody() {
  for(int i=0; i<noteSequenceLength; i++) {
    playNote(noteSequence[i]);
  }
}

void playNote(int thisNote) {
  if(thisNote >= 0 && thisNote <= 7) {
    // note is valid, play note
    bigServo.write(notePositions[thisNote]);
    delay(preHitDelay);
    littleServo.write(beaterHitAngle);
    delay(postHitDelay);
    littleServo.write(beaterRestAngle);    
  } else {
    // delay instead of playing note
    delay(preHitDelay + postHitDelay);
  }
  delay(tempoDelay);
}

void flashLEDs() {
  for(int i=0; i<5; i++) {
    digitalWrite(ledPin, HIGH);
    delay(100);
    digitalWrite(ledPin, LOW);
    delay(100);
  }
  digitalWrite(ledPin, alarmSet);
}

// Function to connect and reconnect as necessary to the MQTT server.
// Should be called in the loop function and it will take care if connecting.
void MQTT_connect() {
  int8_t ret;

  // Stop if already connected.
  if (mqtt.connected()) {
    return;
  }

  Serial.print("Connecting to MQTT... ");

  while ((ret = mqtt.connect()) != 0) { // connect will return 0 for connected
       Serial.println(mqtt.connectErrorString(ret));
       Serial.println("Retrying MQTT connection in 5 seconds...");
       mqtt.disconnect();
       delay(5000);  // wait 5 seconds
  }
  Serial.println("MQTT Connected!");
}
